#! /usr/bin/perl

use Cwd;
use Socket;
use IO::Handle;

sub smtp_result ();
sub usage ();
sub ppsplit ($);
sub SendTo (*@);

($pname, $ppath) = ppsplit ($0);

usage () if (@ARGV < 3);

$DEBUG = 1;
$SENDER = (shift);
$SERVER = (shift);
@RECIPS = @ARGV;
@MAIL = ();
while (<STDIN>) { chomp; s/\r$//g; push @MAIL, $_; }

# get the full qualified domain name of the host (required for the
# HELO-command).
$hostname = `hostname -f`; chomp $hostname; $hostname =~ s/\r$//;

# get the domain name of the host (required for a complete sender's address).
$domname = `hostname -d`; chomp $domname; $domname =~ s/\r$//;
# Add the sender domain if missing.
$SENDER .= "\@$domname" unless ($SENDER =~ /\@[^\@]+$/);

# get the TCP/IP-data
$port = getservbyname ('smtp', 'tcp') || die "$pname: no valid port!\n";
$proto = getprotobyname('tcp');
$iaddr = gethostbyname ($SERVER);
unless (defined $iaddr) {
    $iaddr = inet_aton ($SERVER);
    $name = gethostbyaddr ($iaddr, AF_INET);
    unless (defined $name) {
	die "$pname: invalid <server> argument: `$SERVER'.\n";
    }
    $SERVER = $name;
}
$paddr = sockaddr_in ($port, $iaddr);

# Create a SMTP-connection to <server>
socket (SOCK, PF_INET, SOCK_STREAM, $proto) || die "$pname(socket): $!\n";
connect (SOCK, $paddr) || die "$pname(connect): $!\n";

# Begin with the transmission of the e-mail.
#
($stat, $msg) = smtp_result ();
if (int ($stat / 100) != 2) { die "$pname(SMTP): $msg\n"; }

SOCK->autoflush (1);

SendTo SOCK, "HELO $hostname\r\n";
($stat, $msg) = smtp_result ();
if (int ($stat / 100) != 2) { die "$pname(SMTP): $msg\n"; }

$| = 1;
$pc = 0; $lc = 0; $sr = 0;
print "Sending message to (`" . (join "', `", @RECIPS) . "') ... ";
SendTo SOCK, "MAIL FROM:<$SENDER>\r\n";
($stat, $msg) = smtp_result ();
if (int ($stat / 100) != 2) {
    print STDERR "(MAIL) failed: $msg\n";
    sleep 1;
    SendTo SOCK, "QUIT\r\n"; ($stat, $msg) = smtp_result ();
    close SOCK;
    exit 1;
}
for $recip (@RECIPS) {
    SendTo SOCK, "RCPT TO:<$recip\@$SERVER>\r\n";
    ($stat, $msg) = smtp_result ();
    if (int ($stat / 100) != 2) {
	print STDERR "(RCPT) failed for `$recip': $msg\n";
	next;
    }
    $sr += 1;
}
unless ($sr) {
    SendTo SOCK, "QUIT\r\n"; ($stat, $msg) = smtp_result ();
    close SOCK;
    print STDERR "no valid recipient found ... exiting!\n";
    exit 1;
}
SendTo SOCK, "DATA\r\n";
($stat, $msg) = smtp_result ();
if (int ($stat / 100) != 3) {
    SendTo SOCK, "QUIT\r\n"; ($stat, $msg) = smtp_result ();
    close SOCK;
    print STDERR "(DATA) failed: $msg\n";
    print STDERR "... exiting!\n";
    exit 1;
}

for $_ (@MAIL) {
    if ($_ eq '.') { SendTo SOCK, "."; }
    SendTo SOCK, "$_\r\n";
}

SendTo SOCK, ".\r\n";
($stat, $msg) = smtp_result ();
if (int ($stat / 100) != 2) {
    SendTo SOCK, "QUIT\r\n"; ($stat, $msg) = smtp_result ();
    close SOCK;
    print STDERR "(DATA) failed: $msg\n";
    print STDERR "... exiting!\n";
    exit 1;
}

SendTo SOCK, "QUIT\r\n"; ($stat, $msg) = smtp_result ();
close SOCK;
if (int ($stat / 100) != 2) {
    print STDERR "WARNING! QUIT failed: $msg\n";
}
print " done.\n";

print "+++ $pname complete. +++\n";

exit 0;

sub smtp_result () {
    my ($stat, $msg);
    while (<SOCK>) {
	chomp; s/\r$//; $msg = $_;
	if ($DEBUG) { print STDERR "$_\n"; }
	last if ((substr $msg, 3, 1) eq ' ');
    }
    $stat = (substr $msg, 0, 3) + 0;
    ($stat, $msg)
}

sub usage () {
    die "Usage: $pname <sender> <server> <recip-user-part>...\n";
}

sub ppsplit ($) {
    use Cwd;
    my @pp = split '/', $_[0];
    my (@res, $x);
    my $pn = pop @pp;
    unless (@pp && (length $pp[0]) == 0) { unshift @pp, (split '/', (cwd ())); }
    @res = ();
    while (@pp) {
	$x = shift @pp;
	if ($x eq '..') {
	    pop @res if (@res > 1);
	} elsif ($x ne '.') {
	    push @res, $x;
	}
    }
    ($pn, join ('/', @res))
}

sub SendTo (*@) {
    local *FH = (shift @_);
    my @args = @_;
    if ($DEBUG) { print STDERR @args; }
    print FH @args;
}
