/* cursor.c
**
** implement a simple database cursor concept for a GDBM based database
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include "ifconf.h"
#include "dbif_priv.h"
#include "prefs.h"
#include "tools.h"
#include "exithl.h"
#include "vxerror.h"
#include "selcc.h"
#include "calc_fsize.h"

static int seq_num = 0;

#if 0
/* - siehe an der Aufrufstelle von `link_cl' */
static void link_cl (DBIF_CURSOR cu);
static void unlink_cl (DBIF_CURSOR cu);
#endif

/* dbif_select (db, key, ccflags)
**     selects all records from the database which match the condition
**     constructed fromm the key `key' and the condition flags `ccflags'.
**     the result is a cursor which allows a sequential access to the
**     resulting records.
*/
/*DBIF_CURSOR dbif_select (DBIF_FILE db, char *key, rrel_t ccfunc)*/
DBIF_CURSOR dbif_select (DBIF_FILE db, char *key, seltype sel)
{
    /* Was mu ich tun?
     * - Lesen des Datensatzes zu dem Argument `key' aus der Datenbank
     * - Konstruktion des Dateinamens fr die Ergebnisdatei
     * - Erzeugung der Ergebnisdatei
     * - Sequentielles Lesen aller Datenstze aus der Datenbank;
     *   bertragen der zu der Bedingung passenden Datenstze in die
     *   Ergebnisdatei; nach jedem Datensatz ein `xSet', damit eine
     *   Indexierung mglich ist.
     * - Umschalten der Ergebnisdatei in den Lesemodus
     * - Erzeugung der Cursor-Datenstruktur; wenn dies unmglich ist, dann
     *   Freigabe aller Resourcen.
     */
    DBIF_CURSOR dbc = (DBIF_CURSOR) 0;
    char *resn, fnbuf[49], *p;
    size_t maxsize = 0, dbcl;
    dbif_rec *rec = (dbif_rec *) 0;
    datum dkey, drec;
    xFile resf;
    /* 1. Lesen des Datensatzes zu `key' aus der Datenbank */
    if (key && !(rec = dbif_get (db, key))) { return dbc; }
    /* 2. Konstruktion des Dateinamens fr die Ergebnisdatei */
    sprintf (fnbuf, "select-%d.%d", getpid (), seq_num++);
    if (prefs.listdir[strlen (prefs.listdir) - 1] == '/') {
	resn = concat (prefs.listdir, fnbuf, NULL);
    } else {
	resn = concat (prefs.listdir, "/", fnbuf, NULL);
    }
    if (!resn) { dbif_free (rec); if_errno = DBIF_ENOMEM; return dbc; }
    /* Der Name der Ergebnisdatei ist: <LIST>/select-<pid>.<seq> */
    /* 3. Erzeugung der Ergebnisdatei */
    if (!(resf = xOpen (resn, "ux", 0600))) {
	xUnlink (resn); gfree (resn);
	dbif_free (rec);
	if_errno = DBIF_CCFAIL;
	return dbc;
    }
    /* 24.5.2000 by bj:
    ** Ich mache offensichtlich stndig Fehler in der Listenverwaltung;
    ** auerdem ist nicht einzusehen, warum Dateien, die nur solange bestehen
    ** sollen, wie sie geffnet sind, lnger als ntig einen Namen haben
    ** sollen; schlielich lt die Unix-Architektur das Lschen einer
    ** geffneten Datei zu. Diese Datei wird erst dann wirklich gelscht, wenn
    ** das letzte offene File Handle auf diese Datei verschwunden ist; trotzdem
    ** kann der Verzeichniseintrag ja schon vorher verschwinden. (Unter W**doof
    ** wrde das ganze nicht funktionieren - aber wer verwendet schon W**doof
    ** zusammen mit `VxMail'?)
    */
    xUnlink (resn);
    /* 4. Lesen aller Datenstze aus der Datenbank; bertragen der passenden
    **    Datenstze in die Ergebnisdatei
    */
    dkey = gdbm_firstkey (db->dbf);
    while (dkey.dptr) {
	datum dkey1;
	size_t eml;
	int nusers, maxusers;
	dbif_rec *irec;
	char *ikey, *p, buf[sizeof(size_t) + sizeof(off_t) + 2*sizeof(int) + 1];
	/* Hole den zu `dkey' gehrenden Datensatz aus der Datenbank */
	drec = gdbm_fetch (db->dbf, dkey);
	if (!drec.dptr) {
	    /* Lesen fehlgeschlagen; Resourcen freigeben und raus! */
	    free (dkey.dptr);
	    xClose (resf); xUnlink (resn); gfree (resn);
	    dbif_free (rec);
	    return dbc;
	}
	/* Erzeuge einen zur DB-Schnittstelle passenden Schlssel aus `dkey' */
	eml = dkey.dsize;
	if (!(ikey = (char *) malloc (eml + 1))) {
	    /* Fehlgeschlagen; Resourcen freigeben und raus! */
	    free (dkey.dptr);
	    xClose (resf); xUnlink (resn); gfree (resn);
	    dbif_free (rec);
	    if_errno = DBIF_ENOMEM;
	    return dbc;
	}
	memcpy (ikey, dkey.dptr, eml);
	ikey[eml] = '\0';
	/* Erzeuge einen zur DB-Schnittstelle passenden Datensatz aus `drec' */
	if (!(irec = dbif_mkreca (drec.dptr))) {
	    /* Fehlgeschlagen; Resourcen freigeben und raus! */
	    free (ikey); free (drec.dptr); free (dkey.dptr);
	    xClose (resf); xUnlink (resn); gfree (resn);
	    dbif_free (rec);
	    return dbc;
	}
	p = buf;
	set_size(irec->msg_limit, p); p += sizeof(size_t);
	set_offs(irec->fld_limit, p); p += sizeof(off_t);
	/* bertrage die Benutzerbeschrnkungen einer Domain in den Puffer
	** fr die Ausgabe.
	*/
	if (!dbif_getlim (irec, &nusers, &maxusers)) {
	    set_int (nusers, p); p += sizeof(int);
	    set_int (maxusers, p); p += sizeof(int);
	}
	/* NEW: 08/23/2000
	** In order to be able to get the record type later without knowing
	** the exact structure of the data (depends on the record type), the
	** record type MUST be inserted as the last item of the record!
	*/
	*p++ = dbif_rectype(irec);
	/* Wenn der Datensatz den Bedingungen gengt (<- ccCheck), dann
	** bertrage ihn - mitsamt Schlssel - in die Datei
	*/
	if ((selcc (sel)) (key, rec, ikey, irec))
	{
	    /* NEW: 08/23/2000
	    ** when assuming more than one LLTO-targets, a simple
	    ** `strlen (irec->folder) + 1' is no longer enough for
	    ** calculating the folder length. i introduced a new header
	    ** file containing an inline function for this purpose instead.
	    */
	    size_t pwl, fll, tll;
		   pwl = strlen (irec->passwd) + 1;
		   fll = calc_fsize (dbif_rectype(irec), irec->folder);
		   tll = p - buf;
	    if (xWrite (resf, ikey, eml + 1) < 0
	    ||  xWrite (resf, irec->passwd, pwl) < 0
	    ||  xWrite (resf, irec->folder, fll) < 0
/*	    ||  xWrite (resf, irec->flags, tll) < 0) {*/
	    ||  xWrite (resf, buf, tll) < 0) {
		xClose (resf); xUnlink (resn); gfree (resn);
		free (ikey); free (irec); free (drec.dptr); free (dkey.dptr);
		if_errno = DBIF_WRFAIL;
		return dbc;
	    }
	    /* Merke die Maximallnge der bisher geschriebenen Datenstze
	    ** fr den Lesepuffer des Cursors
	    */
	    if (maxsize < eml + 1 + pwl + fll + tll) {
		maxsize = eml + 1 + pwl + fll + tll;
	    }
	    /* Trage ein, da der aktuelle Datensatz zuende ist */
	    if (xSet (resf) < 0) { 
		xClose (resf); xUnlink (resn); gfree (resn);
		free (ikey); free (irec); free (drec.dptr); free (dkey.dptr);
		if_errno = DBIF_WRFAIL;
		return dbc;
	    }
	}
	/* Gib die alloziierten Strukturen wieder frei */
	free (ikey); free (irec); free (drec.dptr);
	/* Hole den Schlssel fr den nchsten Datensatz */
	dkey1 = gdbm_nextkey (db->dbf, dkey);
	/* Gib den Speicherplatz fr den alten Schlssel frei und bertrage
	** den Schlssel fr den nchsten Datensatz an die entsprechende
	** Variable.
	*/
	free (dkey.dptr); dkey = dkey1;
    }
    /* 5. Umschalten der Ergebnisdatei in den Lesemodus */
    xReSet (resf);
    /* 6. Erzeugung der Datenstruktur fr den `cursor' */
    dbcl = sizeof(struct _DBIF_CURSOR_S) + strlen(resn) + maxsize + 2;
    if (!(dbc = (DBIF_CURSOR) malloc(dbcl))) {
	/* Fehlgeschlagen! Mache auch alles andere wieder rckgngig */
	xClose (resf); xUnlink (resn); gfree (resn);
	dbif_free (rec);
	if_errno = DBIF_ENOMEM;
	return dbc;
    }
    /* Erst dann, wenn wir hier angekommen sind, ist alles erfolgreich
    ** verlaufen; belege nun den neu alloziierten - aber noch leeren
    ** Speicherbereich mit den ermittelten Werten und gib ihn danach zurck.
    */
    p = (char *) dbc + sizeof(struct _DBIF_CURSOR_S);
    /* Den Dateinamen der Cursordatei bentige ich fr die sptere Freigabe
    ** der Resourcen, d.h das Lschen dieser Datei; dieser Name wird direkt
    ** hinter der Struktur eingetragen.
    */
    dbc->res_name = p; strcpy (p, resn); p += strlen(resn) + 1;
    /* Das `xFile'-Handle wird in die Struktur eingetragen */
    dbc->results = resf;
    /* Der Zeiger auf den Lesepuffer - der direkt nach dem Namen der
    ** Cursordatei beginnt - wird eingetragen
    */
    dbc->buf = p;
    /* Die Puffergre wird eingetragen */
    dbc->bufsize = maxsize;
    /* Die Nummer des ersten Datensatzes in der Cursordatei wird eingetragen */
    dbc->recno = 0;
    /* Ich hatte weiter oben den Datensatz zu dem Argument `key' eingelesen;
    ** der Speicherplatz fr diesen mu nun wieder freigegeben werden.
    */
    dbif_free (rec);
#if 0
/* 24.5.2000 by bj:
** Die Liste - samt `exit'-Handler wird nicht lnger bentigt; warum?
** a) Die temporren Dateien sollen sowieso nur solange existieren, wie sie
**    geffnet sind.
** b) Ich gehe (MUSS) ganz klar von einer UNIX-Architektur aus. In dieser ist
**    das Lschen einer bereits geffneten Datei aber gestattet; zunchst wird
**    nmlich nur der `inode' freigegeben; erst wenn die Datei irgendwann
**    geschlossen wird, dann werden auch die Datenblcke freigegeben.
*/
    /* Noch in die Liste eintragen (und evtl. den `exit'-Handler installieren).
    */
    link_cl (dbc);
#endif
    /* So - alles geschafft! */
    return dbc;
}

/* key = dbif_next (dbc, &userrec)
**     reads the next record the database cursor `dbc' points to.
*/
char *dbif_next (DBIF_CURSOR dbc, dbif_rec **userrec)
{
    int ec = xSeek (dbc->results, dbc->recno), rt, nusers, maxusers;
    char *em_addr = (char *) 0, *p;
    dbif_rec *rec;
    size_t eml, fll;
    ssize_t readlen;
    char *ema, *pwd, *fld;
    size_t msgmax; off_t fldmax;
    if (ec < 0) { if_errno = DBIF_NOERR; return em_addr; }
    if ((readlen = xRead (dbc->results, dbc->buf, dbc->bufsize)) <= 0) {
	if_errno = DBIF_RDFAIL;
	return em_addr;
    }
    /* ec ist eigentlich die Anzahl gelesener Bytes (oder < 0 fr einen Fehler);
    ** Ich bentige jedoch nur den Fehler, deshalb der Variablenname:
    ** `ec' steht fr (e)rror (c)ode. Ein Wert ec = 0 steht brigens fr ein
    ** Dateiende.
    */
    /* NEW: 08/23/2000
    ** The record type is the last item in the buffer!
    */
    rt = (int) dbc->buf[readlen - 1];
    p = dbc->buf;        ema = p; eml = strlen (p);
    p += eml + 1;        pwd = p;
    p += strlen (p) + 1; fld = p; fll = calc_fsize (rt, p);
    p += fll;
    msgmax = get_size (p); p += sizeof(size_t);
    fldmax = get_offs (p); p += sizeof(off_t);
/*    rt = (int) *p++;*/
    if (rt == RT_DOMAIN || rt == RT_NSDOMAIN) {
	nusers = get_int (p); p += sizeof(int);
	maxusers = get_int (p);
	/* 08/28/2000: no changes required, because `dbif_mkrec' recognizes
	**             RT_LLINK data automatically.
	*/
	rec = dbif_mkrec (pwd, fld, msgmax, fldmax, rt, nusers, maxusers);
    } else {
	/* 08/28/2000: no changes required, because `dbif_mkrec' recognizes
	**             RT_LLINK data automatically.
	*/
	rec = dbif_mkrec (pwd, fld, msgmax, fldmax, rt);
    }
    /* Diese Funktion hat zwei Ergebnisse: den Schlssel und den Datensatz;
    ** der Schlssel wird als Funktionswert zurckgegeben; er fungiert
    ** zugleich als Dateiende- oder Fehlerindikator; der Datensatz wird
    ** ber den Parameter `userrec' zurckgegeben.
    */
    if (!rec) { return em_addr; }
    if (!(em_addr = malloc (eml + 1))) {
	free (rec);
	if_errno = DBIF_ENOMEM;
	return em_addr;
    }
    strcpy (em_addr, ema);
    *userrec = rec;
    /* War die Leseoperation erfolgreich, dann wird der Zeiger auf das nchste
    ** Ergebnis von `dbif_select' gesetzt.
    */
    ++dbc->recno;
    if_errno = DBIF_NOERR;
    return em_addr;
}

/* key = dbif_first (dbc, &userrec)
**     resets the database cursor `dbc' to the first result of the
**     `dbif_select' call which created this cursor and call `dbif_next' in
**     order to return this result.
*/
char *dbif_first (DBIF_CURSOR dbc, dbif_rec **userrec)
{
    dbc->recno = 0;
    return dbif_next (dbc, userrec);
}

/* dbif_end (dbc)
**     release all resources which were allocated for the database cursor
**     `dbc'.
*/
int dbif_end (DBIF_CURSOR dbc)
{
    xClose (dbc->results);
#if 0
/* - siehe oben an der Aufrufstele von `link_cl' */
    if (xUnlink (dbc->res_name)) {
	vxerror (0, 0, "unable to delete select-file: %s\n", strerror(errno));
    }
    unlink_cl (dbc);
#endif
    free (dbc);
    if_errno = DBIF_NOERR;
    return 0;
}

typedef struct clist_s *clist_t;
struct clist_s {
    clist_t next;
    DBIF_CURSOR unclosed;
};

#if 0
/* Wird nicht lnger bentigt, da die temporren Dateien jetzt sofort nach
** der Erffnung gelscht werden.
*/
clist_t cursors = (clist_t) 0;

static void unopen (void)
{
    clist_t p = cursors;
    while (cursors) {
	dbif_end (cursors->unclosed);
    }
}

static void link_cl (DBIF_CURSOR cu)
{
    clist_t p;
    if (!is_exithl ("rm_cursors")) { add_exithl ("rm_cursors", unopen); }
    p = galloc (__FILE__, __LINE__, sizeof (struct clist_s));
    p->next = cursors;
    p->unclosed = cu;
    cursors = p;
}

static void unlink_cl (DBIF_CURSOR cu)
{
    clist_t p, q;
    p = cursors;
    if (p->unclosed == cu) {
	cursors = p->next;
	gfree (p);
    } else {
	for (q = p, p = q->next; p != NULL; q = p, p = q->next) {
	    if (p->unclosed == cu) {
		q->next = p->next;
		gfree (p);
		break;
	    }
	}
    }
}
#endif
